
void raster_generic(void *destbase, Bit32s y, const poly_extent *extent, const void *extradata, int threadid) 
{
	const poly_extra_data *extra = (const poly_extra_data *)extradata;
	voodoo_state *v = extra->state;
	stats_block *stats = &v->thread_stats[threadid];
	DECLARE_DITHER_POINTERS;
	Bit32s startx = extent->startx;
	Bit32s stopx = extent->stopx;
	Bit32s iterr, iterg, iterb, itera;
	Bit32s iterz;
	Bit64s iterw, iterw0 = 0, iterw1 = 0;
	Bit64s iters0 = 0, iters1 = 0;
	Bit64s itert0 = 0, itert1 = 0;
	Bit16u *depth;
	Bit16u *dest;
	Bit32s dx, dy;
	Bit32s scry;
	Bit32s x;

	/* determine the screen Y */
	scry = y;
	if (FBZMODE_Y_ORIGIN(extra->r_fbzMode))
		scry = (v->fbi.yorigin - y) & 0x3ff;

	/* compute dithering */
	COMPUTE_DITHER_POINTERS(extra->r_fbzMode, y);

	/* apply clipping */
	if (FBZMODE_ENABLE_CLIPPING(extra->r_fbzMode))
	{
		Bit32s tempclip;

		/* Y clipping buys us the whole scanline */
		if (scry < ((v->reg[clipLowYHighY].u >> 16) & 0x3ff) ||
			scry >= (v->reg[clipLowYHighY].u & 0x3ff))
		{
			stats->pixels_in += stopx - startx;
			stats->clip_fail += stopx - startx;
			return;
		}

		/* X clipping */
		tempclip = (v->reg[clipLeftRight].u >> 16) & 0x3ff;
		if (startx < tempclip)
		{
			stats->pixels_in += tempclip - startx;
			v->stats.total_clipped += tempclip - startx;
			startx = tempclip;
		}
		tempclip = v->reg[clipLeftRight].u & 0x3ff;
		if (stopx >= tempclip)
		{
			stats->pixels_in += stopx - tempclip;
			v->stats.total_clipped += stopx - tempclip;
			stopx = tempclip - 1;
		}
	}

	/* get pointers to the target buffer and depth buffer */
	dest = (Bit16u *)destbase + scry * v->fbi.rowpixels;
	depth = (v->fbi.auxoffs != ~0) ? ((Bit16u *)(v->fbi.ram + v->fbi.auxoffs) + scry * v->fbi.rowpixels) : NULL; 

	/* compute the starting parameters */
	dx = startx - (extra->ax >> 4);
	dy = y - (extra->ay >> 4);
	iterr = extra->startr + dy * extra->drdy + dx * extra->drdx;
	iterg = extra->startg + dy * extra->dgdy + dx * extra->dgdx;
	iterb = extra->startb + dy * extra->dbdy + dx * extra->dbdx;
	itera = extra->starta + dy * extra->dady + dx * extra->dadx;
	iterz = extra->startz + dy * extra->dzdy + dx * extra->dzdx;
	iterw = extra->startw + dy * extra->dwdy + dx * extra->dwdx;
	if (extra->texcount >= 1)
	{
		iterw0 = extra->startw0 + dy * extra->dw0dy +dx * extra->dw0dx;
		iters0 = extra->starts0 + dy * extra->ds0dy + dx * extra->ds0dx;
		itert0 = extra->startt0 + dy * extra->dt0dy + dx * extra->dt0dx;
	}
	if (extra->texcount >= 2)
	{
		iterw1 = extra->startw1 + dy * extra->dw1dy +dx * extra->dw1dx;
		iters1 = extra->starts1 + dy * extra->ds1dy + dx * extra->ds1dx;
		itert1 = extra->startt1 + dy * extra->dt1dy + dx * extra->dt1dx;
	}

	/* loop in X */
	for (x = startx; x < stopx; x++)
	{
		rgb_union iterargb = { 0 };
		rgb_union texel = { 0 };

		/* pixel pipeline part 1 handles depth testing and stippling */
		PIXEL_PIPELINE_BEGIN(v, stats, x, y, extra->r_fbzColorPath, extra->r_fbzMode, 
			iterz, iterw);

		/* run the texture pipeline on TMU1 to produce a value in texel */
		/* note that they set LOD min to 8 to "disable" a TMU */
		if (extra->texcount >= 2 && v->tmu[1].lodmin < (8 << 8))
			TEXTURE_PIPELINE(&v->tmu[1], x, dither4, extra->r_textureMode1, texel,
			v->tmu[1].lookup, extra->lodbase1,
			iters1, itert1, iterw1, texel);

		/* run the texture pipeline on TMU0 to produce a final */
		/* result in texel */
		/* note that they set LOD min to 8 to "disable" a TMU */
		if (extra->texcount >= 1 && v->tmu[0].lodmin < (8 << 8))
/*
			TEXTURE_PIPELINE(&v->tmu[0], x, dither4, extra->r_textureMode0, texel,
			v->tmu[0].lookup, extra->lodbase0,
			iters0, itert0, iterw0, texel);
*/
			do
			{
				tmu_state *tt=&v->tmu[0];
				Bit32s xx=x;
				int texmode=extra->r_textureMode0;
				rgb_union cother=texel;
				rgb_t *lookup=v->tmu[0].lookup;
				Bit32s lodbase=extra->lodbase0;
				Bit64s iters=iters0;
				Bit64s itert=itert0;
				Bit64s iterw=iterw0;
				rgb_union result;


				Bit32s blendr, blendg, blendb, blenda;
				Bit32s tr, tg, tb, ta;
				Bit32s oow, s, t, lod, ilod;
				Bit32s smax, tmax;
				Bit32u texbase;
				rgb_union c_local;

				/* determine the S/T/LOD values for this texture */
				if (TEXMODE_ENABLE_PERSPECTIVE(texmode))
				{
					oow = fast_reciplog((iterw), &lod);
					s = ((Bit64s)oow * (iters)) >> 29;
					t = ((Bit64s)oow * (itert)) >> 29;
					lod += (lodbase);
				}
				else
				{
					s = (iters) >> 14;
					t = (itert) >> 14;
					lod = (lodbase);
				}

				/* clamp W */
				if (TEXMODE_CLAMP_NEG_W(texmode) && (iterw) < 0)
					s = t = 0;

				/* clamp the LOD */
				lod += tt->lodbias;
				if (TEXMODE_ENABLE_LOD_DITHER(texmode))
					lod += dither4[xx & 3] << 4;
				if (lod < tt->lodmin)
					lod = tt->lodmin;
				if (lod > tt->lodmax)
					lod = tt->lodmax;

				/* now the LOD is in range; if we don't own this LOD, take the next one */
				ilod = lod >> 8;
				if (!((tt->lodmask >> ilod) & 1))
					ilod++;

				/* fetch the texture base */
				texbase = tt->lodoffset[ilod];

				/* compute the maximum s and t values at this LOD */
				smax = tt->wmask >> ilod;
				tmax = tt->hmask >> ilod;

				/* determine whether we are point-sampled or bilinear */
				if ((lod == tt->lodmin && !TEXMODE_MAGNIFICATION_FILTER(texmode)) ||
					(lod != tt->lodmin && !TEXMODE_MINIFICATION_FILTER(texmode)))
				{
					/* point sampled */

					Bit32u texel0;

					/* adjust S/T for the LOD and strip off the fractions */
					s >>= ilod + 18;
					t >>= ilod + 18;

					/* clamp/wrap S/T if necessary */
					if (TEXMODE_CLAMP_S(texmode))
						CLAMP(s, 0, smax);
					if (TEXMODE_CLAMP_T(texmode))
						CLAMP(t, 0, tmax);
					s &= smax;
					t &= tmax;
					t *= smax + 1;

					/* fetch texel data */
					if (TEXMODE_FORMAT(texmode) < 8)
					{
						texel0 = *(Bit8u *)&tt->ram[(texbase + t + s) & tt->mask];
						c_local.u = lookup[texel0];
					}
					else
					{
						texel0 = *(Bit16u *)&tt->ram[(texbase + 2*(t + s)) & tt->mask];
						if (TEXMODE_FORMAT(texmode) >= 10 && TEXMODE_FORMAT(texmode) <= 12)
							c_local.u = lookup[texel0];
						else
							c_local.u = (lookup[texel0 & 0xff] & 0xffffff) |
							((texel0 & 0xff00) << 16);
					}
				}
				else
				{
					/* bilinear filtered */

					Bit32u texel0, texel1, texel2, texel3;
					Bit32u sfrac, tfrac;
					Bit32s s1, t1;

					/* adjust S/T for the LOD and strip off all but the low 8 bits of */
					/* the fraction */
					s >>= ilod + 10;
					t >>= ilod + 10;

					/* also subtract 1/2 texel so that (0.5,0.5) = a full (0,0) texel */
					s -= 0x80;
					t -= 0x80;

					/* extract the fractions */
					sfrac = s & tt->bilinear_mask;
					tfrac = t & tt->bilinear_mask;

					/* now toss the rest */
					s >>= 8;
					t >>= 8;
					s1 = s + 1;
					t1 = t + 1;

					/* clamp/wrap S/T if necessary */
					if (TEXMODE_CLAMP_S(texmode))
					{
						CLAMP(s, 0, smax);
						CLAMP(s1, 0, smax);
					}
					if (TEXMODE_CLAMP_T(texmode))
					{
						CLAMP(t, 0, tmax);
						CLAMP(t1, 0, tmax);
					}
					s &= smax;
					s1 &= smax;
					t &= tmax;
					t1 &= tmax;
					t *= smax + 1;
					t1 *= smax + 1;

					/* fetch texel data */
					if (TEXMODE_FORMAT(texmode) < 8)
					{
						texel0 = *(Bit8u *)&tt->ram[(texbase + t + s) & tt->mask];
						texel1 = *(Bit8u *)&tt->ram[(texbase + t + s1) & tt->mask];
						texel2 = *(Bit8u *)&tt->ram[(texbase + t1 + s) & tt->mask];
						texel3 = *(Bit8u *)&tt->ram[(texbase + t1 + s1) & tt->mask];
						texel0 = lookup[texel0];
						texel1 = lookup[texel1];
						texel2 = lookup[texel2];
						texel3 = lookup[texel3];
					}
					else
					{
						texel0 = *(Bit16u *)&tt->ram[(texbase + 2*(t + s)) & tt->mask];
						texel1 = *(Bit16u *)&tt->ram[(texbase + 2*(t + s1)) & tt->mask];
						texel2 = *(Bit16u *)&tt->ram[(texbase + 2*(t1 + s)) & tt->mask];
						texel3 = *(Bit16u *)&tt->ram[(texbase + 2*(t1 + s1)) & tt->mask];
						if (TEXMODE_FORMAT(texmode) >= 10 && TEXMODE_FORMAT(texmode) <= 12)
						{
							texel0 = lookup[texel0];
							texel1 = lookup[texel1];
							texel2 = lookup[texel2];
							texel3 = lookup[texel3];
						}
						else
						{
							texel0 = (lookup[texel0 & 0xff] & 0xffffff) | 
								((texel0 & 0xff00) << 16);
							texel1 = (lookup[texel1 & 0xff] & 0xffffff) | 
								((texel1 & 0xff00) << 16);
							texel2 = (lookup[texel2 & 0xff] & 0xffffff) | 
								((texel2 & 0xff00) << 16);
							texel3 = (lookup[texel3 & 0xff] & 0xffffff) | 
								((texel3 & 0xff00) << 16);
						}
					}

					/* weigh in each texel */
					c_local.u = rgba_bilinear_filter(texel0, texel1, texel2, texel3, sfrac, tfrac);
				}

				/* select zero/other for RGB */
				if (!TEXMODE_TC_ZERO_OTHER(texmode))
				{
					tr = cother.rgb.r;
					tg = cother.rgb.g;
					tb = cother.rgb.b;
				}
				else
					tr = tg = tb = 0;

				/* select zero/other for alpha */
				if (!TEXMODE_TCA_ZERO_OTHER(texmode))
					ta = cother.rgb.a;
				else
					ta = 0;

				/* potentially subtract c_local */
				if (TEXMODE_TC_SUB_CLOCAL(texmode))
				{
					tr -= c_local.rgb.r;
					tg -= c_local.rgb.g;
					tb -= c_local.rgb.b;
				}
				if (TEXMODE_TCA_SUB_CLOCAL(texmode))
					ta -= c_local.rgb.a;

				/* blend RGB */
				switch (TEXMODE_TC_MSELECT(texmode))
				{
				default:/* reserved */
				case 0:/* zero */
					blendr = blendg = blendb = 0;
					break;

				case 1:/* c_local */
					blendr = c_local.rgb.r;
					blendg = c_local.rgb.g;
					blendb = c_local.rgb.b;
					break;

				case 2:/* a_other */
					blendr = blendg = blendb = cother.rgb.a;
					break;

				case 3:/* a_local */
					blendr = blendg = blendb = c_local.rgb.a;
					break;

				case 4:/* LOD (detail factor) */
					if (tt->detailbias <= lod)
						blendr = blendg = blendb = 0;
					else
					{
						blendr = (((tt->detailbias - lod) << tt->detailscale) >> 8);
						if (blendr > tt->detailmax)
							blendr = tt->detailmax;
						blendg = blendb = blendr;
					}
					break;

				case 5:/* LOD fraction */
					blendr = blendg = blendb = lod & 0xff;
					break;
				}

				/* blend alpha */
				switch (TEXMODE_TCA_MSELECT(texmode))
				{
				default:/* reserved */
				case 0:/* zero */
					blenda = 0;
					break;

				case 1:/* c_local */
					blenda = c_local.rgb.a;
					break;

				case 2:/* a_other */
					blenda = cother.rgb.a;
					break;

				case 3:/* a_local */
					blenda = c_local.rgb.a;
					break;

				case 4:/* LOD (detail factor) */
					if (tt->detailbias <= lod)
						blenda = 0;
					else
					{
						blenda = (((tt->detailbias - lod) << tt->detailscale) >> 8);
						if (blenda > tt->detailmax)
							blenda = tt->detailmax;
					}
					break;

				case 5:/* LOD fraction */
					blenda = lod & 0xff;
					break;
				}

				/* reverse the RGB blend */
				if (!TEXMODE_TC_REVERSE_BLEND(texmode))
				{
					blendr ^= 0xff;
					blendg ^= 0xff;
					blendb ^= 0xff;
				}

				/* reverse the alpha blend */
				if (!TEXMODE_TCA_REVERSE_BLEND(texmode))
					blenda ^= 0xff;

				/* do the blend */
				tr = (tr * (blendr + 1)) >> 8;
				tg = (tg * (blendg + 1)) >> 8;
				tb = (tb * (blendb + 1)) >> 8;
				ta = (ta * (blenda + 1)) >> 8;

				/* add clocal or alocal to RGB */
				switch (TEXMODE_TC_ADD_ACLOCAL(texmode))
				{
				case 3:/* reserved */
				case 0:/* nothing */
					break;

				case 1:/* add c_local */
					tr += c_local.rgb.r;
					tg += c_local.rgb.g;
					tb += c_local.rgb.b;
					break;

				case 2:/* add_alocal */
					tr += c_local.rgb.a;
					tg += c_local.rgb.a;
					tb += c_local.rgb.a;
					break;
				}

				/* add clocal or alocal to alpha */
				if (TEXMODE_TCA_ADD_ACLOCAL(texmode))
					ta += c_local.rgb.a;

				/* clamp */
				result.rgb.r = (tr < 0) ? 0 : (tr > 0xff) ? 0xff : tr;
				result.rgb.g = (tg < 0) ? 0 : (tg > 0xff) ? 0xff : tg;
				result.rgb.b = (tb < 0) ? 0 : (tb > 0xff) ? 0xff : tb;
				result.rgb.a = (ta < 0) ? 0 : (ta > 0xff) ? 0xff : ta;

				/* invert */
				if (TEXMODE_TC_INVERT_OUTPUT(texmode))
					result.u ^= 0x00ffffff;
				if (TEXMODE_TCA_INVERT_OUTPUT(texmode))
					result.rgb.a ^= 0xff;

				texel=result;
			} while (0);

		/* colorpath pipeline selects source colors and does blending */
		CLAMPED_ARGB(iterr, iterg, iterb, itera, extra->r_fbzColorPath, iterargb);
		COLORPATH_PIPELINE(v, stats, extra->r_fbzColorPath, extra->r_fbzMode, extra->r_alphaMode, texel,
			iterz, iterw, iterargb);

		/* pixel pipeline part 2 handles fog, alpha, and final output */
		PIXEL_PIPELINE_END(v, stats, dither, dither4, dither_lookup, x, dest, depth, 
			extra->r_fbzMode, extra->r_fbzColorPath, extra->r_alphaMode, extra->r_fogMode,
			iterz, iterw, iterargb);

		/* update the iterated parameters */
		iterr += extra->drdx;
		iterg += extra->dgdx;
		iterb += extra->dbdx;
		itera += extra->dadx;
		iterz += extra->dzdx;
		iterw += extra->dwdx;
		if (extra->texcount >= 1)
		{
			iterw0 += extra->dw0dx;
			iters0 += extra->ds0dx;
			itert0 += extra->dt0dx;
		}
		if (extra->texcount >= 2)
		{
			iterw1 += extra->dw1dx;
			iters1 += extra->ds1dx;
			itert1 += extra->dt1dx;
		}
	}
}